@@ -1,4 +1,4 @@
-// Copyright 2017 Lars Kroll. See the LICENSE
+// Copyright 2017-2020 Lars Kroll. See the LICENSE
 // file at the top-level directory of this distribution.
 //
 // Licensed under the MIT license
@@ -6,13 +6,25 @@
 // This file may not be copied, modified, or distributed
 // except according to those terms.
 
+//! The core traits and reusable functions of this crate.
+
 use std::fmt::Debug;
 
+/// A minimal trait for task executors.
+///
+/// This is mostly useful as a narrowed view for dynamic executor references.
+pub trait CanExecute {
+    /// Executes the function `job` on the `Executor`.
+    ///
+    /// This is the same as [execute](Executor::execute), but already boxed up.
+    fn execute_job(&self, job: Box<dyn FnOnce() + Send + 'static>);
+}
+
 /// A common trait for task executors.
 ///
 /// All implementations need to allow cloning to create new handles to
 /// the same executor, and they need to be safe to pass to threads.
-pub trait Executor: Clone + Send {
+pub trait Executor: CanExecute + Clone + Send {
     /// Executes the function `job` on the `Executor`.
     ///
     /// # Examples
@@ -33,7 +45,10 @@ pub trait Executor: Clone + Send {
     /// ```
     fn execute<F>(&self, job: F)
     where
-        F: FnOnce() + Send + 'static;
+        F: FnOnce() + Send + 'static,
+    {
+        self.execute_job(Box::new(job));
+    }
 
     /// Shutdown the `Executor` without waiting.
     ///
@@ -53,9 +68,9 @@ pub trait Executor: Clone + Send {
     /// let executor2 = executor.clone();
     /// executor.execute(|| println!("Hello!"));
     /// executor.execute(move || {
-    /// 		println!("Shutting down");
-    ///		executor2.shutdown_async();
-    ///	});
+    ///     println!("Shutting down");
+    ///     executor2.shutdown_async();
+    /// });
     /// std::thread::sleep(std::time::Duration::from_secs(1)); // or wait with a barrier
     /// executor.execute(|| println!("doesn't work!"));
     /// ```
