@@ -1,4 +1,4 @@
-// Copyright 2017 Lars Kroll. See the LICENSE
+// Copyright 2017-2020 Lars Kroll. See the LICENSE
 // file at the top-level directory of this distribution.
 //
 // Licensed under the MIT license
@@ -6,7 +6,10 @@
 // This file may not be copied, modified, or distributed
 // except according to those terms.
 
-//! A simple `Executor` that simply runs tasks on the current thread.
+//! A simple `Executor` that simply runs tasks immediately on the current thread.
+//!
+//! This is mostly useful to work with APIs that require an `Executor`,
+//! even if "normal" stack-based execution is actually desired.
 //!
 //! # Examples
 //!
@@ -23,18 +26,32 @@
 //! exec.execute(|| println!("bar"));
 //! exec.shutdown();
 //! ```
-
+//!
+//! # Note
+//!
+//! If you use [try_execute_locally](try_execute_locally) from within a job closure,
+//! it will be the same as running recursively, so you may run of out stack space, eventually.
+use super::*;
 use std::sync::atomic::{AtomicBool, Ordering};
 use std::sync::Arc;
 
+/// A handle to the [run_now](run_now) executor
+///
+/// See module level documentation for usage information.
 #[derive(Clone, Debug)]
 pub struct RunNowExecutor {
     active: Arc<AtomicBool>,
 }
 
-use super::*;
+struct ThreadLocalRunNow;
+impl CanExecute for ThreadLocalRunNow {
+    fn execute_job(&self, job: Box<dyn FnOnce() + Send + 'static>) {
+        job();
+    }
+}
 
 impl RunNowExecutor {
+    /// Create a new [run_now](run_now) executor
     pub fn new() -> RunNowExecutor {
         RunNowExecutor {
             active: Arc::new(AtomicBool::new(true)),
@@ -50,13 +67,26 @@ impl Default for RunNowExecutor {
     }
 }
 
+impl CanExecute for RunNowExecutor {
+    fn execute_job(&self, job: Box<dyn FnOnce() + Send + 'static>) {
+        if self.active.load(Ordering::SeqCst) {
+            job();
+        } else {
+            warn!("Ignoring job as pool is shutting down.");
+        }
+    }
+}
+
 impl Executor for RunNowExecutor {
+    // override default implementation to avoid boxing at all
     fn execute<F>(&self, job: F)
     where
         F: FnOnce() + Send + 'static,
     {
         if self.active.load(Ordering::SeqCst) {
+            set_local_executor(ThreadLocalRunNow);
             job();
+            unset_local_executor();
         } else {
             warn!("Ignoring job as pool is shutting down.");
         }
@@ -100,6 +130,14 @@ mod tests {
         crate::tests::test_defaults::<RunNowExecutor>(LABEL);
     }
 
+    // Will stack overflow.
+    // #[test]
+    // #[should_panic]
+    // fn test_local() {
+    //     let exec = RunNowExecutor::new();
+    //     crate::tests::test_local(exec, LABEL);
+    // }
+
     #[test]
     fn run_tasks() {
         let _ = env_logger::try_init();
